""" XML utilities functions
"""
from xml_utils.xsd_tree.operations.namespaces import (
    get_target_namespace,
    get_namespaces,
)
from xml_utils.xsd_tree.xsd_tree import XSDTree


def get_xpath_from_dot_notation(dot_notation_path):
    """Transform MongoDB dot notation to XPath

    Params:
        dot_notation_path:

    Returns:

    """
    dot_notation_elements = [
        "{0}:%s" % path if not path.startswith("@") else path
        for path in dot_notation_path.split(".")
    ]

    xpath = "/%s" % "/".join(dot_notation_elements)
    return xpath


def get_xpath_with_target_namespace(xpath, xsd_string):
    """Adds target namespace to a given XPath

    Params:
        xpath:
        xsd_string:

    Returns:

    """
    namespaces = get_target_namespace(
        XSDTree.build_tree(xsd_string), get_namespaces(xsd_string)
    )

    xpath = xpath.format(namespaces[1])

    if namespaces[1] == "":
        xpath = xpath.replace(":", "")
        namespaces = None
    else:
        namespaces = {namespaces[1]: namespaces[0]}

    return xpath, namespaces


def set_value_at_xpath(xml_tree, xpath, value, namespaces):
    """Set value for a given XPath

    Params:
        xml_tree:
        xpath:
        value:
        namespaces:

    Returns:

    """
    try:
        xml_tree.xpath(xpath, namespaces=namespaces)[0].text = value
    except AttributeError:
        xpath_list = xpath.split("/")
        attribute = xpath_list[-1].replace("@", "")
        xpath = "/%s" % "/".join(xpath_list[:-1])

        xml_tree.xpath(xpath, namespaces=namespaces)[0].attrib[attribute] = value


def get_value_at_xpath(xml_tree, xpath, namespaces):
    """Retrieve value in XML given a XPath

    Params:
        xml_tree:
        xpath:
        namespaces:

    Returns:

    """
    xpath_element_list = xml_tree.xpath(xpath, namespaces=namespaces)

    # Assert that we found exactly one element matching the given xpath
    assert len(xpath_element_list) == 1

    try:
        xpath_value = xpath_element_list[0].text
    except AttributeError:  # XPath points to an attribute
        xpath_value = xpath_element_list[0]

    return xpath_value



""" REST views for the query API
"""
import json
from urllib.parse import urljoin


from django.urls import reverse
from rest_framework import status
from rest_framework.response import Response

from core_linked_records_app import settings
from core_main_app.components.data import api as data_api
from core_main_app.rest.data.abstract_views import AbstractExecuteLocalQueryView
from core_main_app.utils.requests_utils.requests_utils import send_get_request


class ExecuteLocalPIDQueryView(AbstractExecuteLocalQueryView):
    def build_query(
        self, query, workspaces=None, templates=None, options=None, title=None
    ):
        """Build the query by adding an extra filter to limit to document with
        PID fields.

        Args:
            query:
            workspaces:
            templates:
            options:
            title:

        Returns:
            Query with additional PID filter
        """
        pid_query = {"$and": [{"dict_content.%s" % settings.PID_XPATH: {"$exists": 1}}]}
        query = super().build_query(query, workspaces, templates, options, title)

        if "$and" in query.keys():
            pid_query["$and"] += query["$and"]
        else:
            pid_query["$and"].append(query)

        return pid_query

    def execute_raw_query(self, raw_query, order_by_field):
        """Execute the raw query in database

        Args:

            raw_query: Query to execute
            order_by_field:

        Returns:
            Results of the query
        """
        pipeline = [
            {"$match": raw_query},
            {"$project": {"pid": "$dict_content.%s" % settings.PID_XPATH}},
        ]
        return data_api.aggregate(pipeline, self.request.user)

    def build_response(self, data_list):
        """Build the list of PIDs

        Args:

            data_list: List of data

        Returns:

            Paginated list of data
        """
        # Send list of PID as JSON
        return Response(
            [data["pid"] for data in data_list],
            status=status.HTTP_200_OK,
        )


if (
    "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
    and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
):
    from core_oaipmh_harvester_app.components.oai_record import api as oai_record_api
    from core_oaipmh_harvester_app.components.oai_registry import (
        api as oai_registry_api,
    )
    from core_explore_oaipmh_app.rest.query.views import ExecuteQueryView

    class ExecuteOaiPmhPIDQueryView(ExecuteQueryView):
        pid_xpath = None

        def build_query(self, query, templates, registries):
            """Build the query by adding an extra filter to limit to document with
            PID fields.

            Args:
                query:
                templates:
                registries:

            Returns:
                Query with additional PID filter
            """
            registry_info = oai_registry_api.get_by_id(json.loads(registries)[0])
            pid_xpath_request = send_get_request(
                urljoin(
                    registry_info.url.replace("/oaipmh/server", ""),
                    reverse("core_linked_records_app_settings"),
                )
            )

            if pid_xpath_request.status_code != status.HTTP_200_OK:
                raise Exception("Impossible to retrieve remote PID settings.")

            self.pid_xpath = pid_xpath_request.json()["xpath"]

            pid_query = {"$and": [{"dict_content.%s" % self.pid_xpath: {"$exists": 1}}]}
            query = super().build_query(query, templates, registries)

            if "$and" in query.keys():
                pid_query["$and"] += query["$and"]
            else:
                pid_query["$and"].append(query)

            return pid_query

        def execute_raw_query(self, raw_query, order_by_field):
            """Execute the raw query in database

            Args:

                raw_query: Query to execute
                order_by_field:

            Returns:
                Results of the query
            """
            if self.pid_xpath is None:
                raise Exception("Undefined PID xpath from remote server.")

            pipeline = [
                {"$match": raw_query},
                {"$project": {"pid": "$dict_content.%s" % self.pid_xpath}},
            ]
            return oai_record_api.aggregate(pipeline, self.request.user)

        def build_response(self, data_list):
            """Build the list of PIDs

            Args:

                data_list: List of data

            Returns:

                Paginated list of data
            """
            # Send list of PID as JSON
            return Response(
                [data["pid"] for data in data_list],
                status=status.HTTP_200_OK,
            )



""" Local resolver API
"""

from core_linked_records_app.settings import PID_XPATH
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_main_app.commons.exceptions import ApiError, DoesNotExist
from core_main_app.components.data import api as data_api


def get_data_by_pid(pid, user):
    """Return data object with the given pid.

    Parameters:
        pid:
        user:

    Returns: data object
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = data_api.execute_query({json_pid_path: pid}, user)
    query_result_length = len(query_result)

    if query_result_length == 0:
        raise DoesNotExist("No result found")
    elif query_result_length != 1:
        raise ApiError("PID must be unique")
    else:
        return query_result[0]


def get_pids_for_data_list(data_id_list, user):
    """Retrieve PID matching the document list provided.

    Args:
        data_id_list:
        user:

    Returns:
    """

    # Retrieve the document passed as input and extra the PID field.
    data_list = data_api.get_by_id_list(data_id_list, user)

    # Build the list of PID from the document and the PID_XPATH
    pid_list = [
        get_dict_value_from_key_list(data["dict_content"], PID_XPATH.split("."))
        for data in data_list
    ]

    # Returns a list of PID available from the list.
    return [pid for pid in pid_list if pid is not None]


def get_pid_for_data(data_id, user):
    """Retrieve PID matching the document ID provided.

    Args:
        data_id:
        user:

    Returns:
    """
    # Retrieve the document passed as input and extra the PID field.
    data = data_api.get_by_id(data_id, user)

    # Return PID value from the document and the PID_XPATH
    return get_dict_value_from_key_list(data["dict_content"], PID_XPATH.split("."))



""" Data Xml renderer for django REST API
"""
from django.http import HttpResponse
from rest_framework import renderers
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_500_INTERNAL_SERVER_ERROR


class DataXmlRenderer(renderers.BaseRenderer):
    media_type = "application/xml"
    format = "xml"
    charset = "utf-8"

    def render(self, data, media_type=None, renderer_context=None):
        """Render the data object by just returning the xml_content field

        Args:
            data:
            media_type:
            renderer_context:

        Returns: xml string
        """

        # check errors
        if "status" in data and data["status"] == "error":
            if data["code"] == HTTP_404_NOT_FOUND:
                return HttpResponse(status=HTTP_404_NOT_FOUND)
            else:
                return HttpResponse(status=HTTP_500_INTERNAL_SERVER_ERROR)
        else:
            return data["xml_content"]



""" Handle.net implementation class
"""
import json
import logging
from base64 import b64encode

from core_linked_records_app.utils.providers import AbstractIdProvider
from core_main_app.utils.requests_utils.requests_utils import (
    send_put_request,
    send_delete_request,
    send_get_request,
)

LOGGER = logging.getLogger(__name__)


class HandleNetSystem(AbstractIdProvider):
    """"""

    registration_api = "api/handles"
    response_code_messages = {
        1: "Successful operation",
        2: "Unexpected handle server error",
        100: "Handle not found",
        101: "Handle already exists",
        102: "Invalid handle",
        200: "Values not found",
        201: "Value already exists",
        202: "Invalid value",
        301: "Server not responsible for handle",
        402: "Authentication needed",
    }

    def _get_message_for_response_code(self, return_code):
        if return_code in self.response_code_messages.keys():
            return self.response_code_messages[return_code]
        else:
            return "Handle code not recognized"

    def _update_response_content(self, response):
        json_response_content = json.loads(response.content)

        json_response_content["message"] = self._get_message_for_response_code(
            json_response_content["responseCode"]
        )

        json_response_content["url"] = "%s/%s" % (
            self.provider_url,
            json_response_content["handle"],
        )

        return json.dumps(json_response_content)

    def encode_token(self, username, password):
        user_pass = "%s:%s" % (username, password)
        return b64encode(user_pass.encode("utf-8")).decode("utf-8")

    def get(self, record):
        """Retrieve an existring handle.net handle.

        Args:
            record:

        Returns:
        """
        response = send_get_request(
            "%s/%s/%s" % (self.provider_url, self.registration_api, record),
            headers={
                "Content-Type": "application/json",
            },
        )

        response._content = self._update_response_content(response)
        return response

    def create(self, prefix, record=None):
        """Create a new handle for a handle.net system.

        Args:
            prefix:
            record:

        Returns:
        """
        # Create request url depending on handle value
        if record is not None:
            request_url = "%s/%s/%s/%s?overwrite=false" % (
                self.provider_url,
                self.registration_api,
                prefix,
                record,
            )
        else:
            request_url = "%s/%s/%s/?overwrite=false&mintNewSuffix=true" % (
                self.provider_url,
                self.registration_api,
                prefix,
            )

        response = send_put_request(
            request_url,
            json.dumps(
                {
                    "values": [
                        {
                            "index": 100,
                            "type": "URL",
                            "data": {
                                "format": "string",
                                "value": "%s/%s/%s" % (self.local_url, prefix, record),
                            },
                        }
                    ]
                }
            ),
            headers={
                "Content-Type": "application/json",
                "Authorization": "Basic %s" % str(self.auth_token),
            },
        )

        if record is None:
            return self.update(response.json()["handle"])
        else:
            response._content = self._update_response_content(response)
            return response

    def update(self, record):
        """Update a handle for a handle.net system.

        Args:
            record:

        Returns:
        """
        response = send_put_request(
            "%s/%s/%s?overwrite=true"
            % (self.provider_url, self.registration_api, record),
            json.dumps(
                {
                    "handle": record,
                    "values": [
                        {
                            "index": 100,
                            "type": "URL",
                            "data": {
                                "format": "string",
                                "value": "%s/%s" % (self.local_url, record),
                            },
                        }
                    ],
                }
            ),
            headers={
                "Content-Type": "application/json",
                "Authorization": "Basic %s" % str(self.auth_token),
            },
        )

        response._content = self._update_response_content(response)
        return response

    def delete(self, record):
        response = send_delete_request(
            "%s/%s/%s" % (self.provider_url, self.registration_api, record),
            headers={"Authorization": "Basic %s" % str(self.auth_token)},
        )

        response._content = self._update_response_content(response)
        return response



""" Local handle model
"""
from django_mongoengine import fields, Document
from core_main_app.commons.regex import NOT_EMPTY_OR_WHITESPACES

from mongoengine import errors as mongoengine_errors
from core_main_app.commons import exceptions


class LocalId(Document):
    """Handle object"""

    record_name = fields.StringField(
        blank=False, unique=True, regex=NOT_EMPTY_OR_WHITESPACES
    )

    @staticmethod
    def get_by_name(record_name):
        try:
            return LocalId.objects.get(record_name=record_name)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))



""" System API for core_linked_records
"""
from core_linked_records_app.settings import PID_XPATH
from core_main_app.commons.exceptions import DoesNotExist, ApiError
from core_main_app.components.data.models import Data


def is_pid_defined_for_document(pid, document_id):
    """Determine if a given PID match the document ID provided.

    Params:
        pid:
        document_id:

    Returns:
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])

    return len(query_result) == 1 and query_result[0].pk == document_id


def is_pid_defined(pid):
    """Determine if a given PID already exists.

    Params:
        pid:

    Returns:
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])

    return len(query_result) == 1


def get_data_by_pid(pid):
    """Return data object with the given pid.

    Parameters:
        pid:
        user:

    Returns: data object
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])
    query_result_length = len(query_result)

    if query_result_length == 0:
        raise DoesNotExist("No result found")
    elif query_result_length != 1:
        raise ApiError("PID must be unique")
    else:
        return query_result[0]



""" Ajax views accessible by users.
"""
import json
from urllib.parse import urljoin

from django.http import JsonResponse
from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.views import APIView

from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.utils.protocols.oauth2 import (
    send_post_request as oauth2_post_request,
    send_get_request as oauth2_get_request,
)
from core_linked_records_app.components.data import api as data_api
from core_main_app.utils.requests_utils.requests_utils import send_get_request
from core_linked_records_app import settings


class RetrieveDataPID(APIView):
    """Retrieve PIDs for a given data IDs."""

    def get(self, request):
        if "data_id" in request.GET:
            return JsonResponse(
                {"pid": data_api.get_pid_for_data(request.GET["data_id"], request.user)}
            )
        elif (
            "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
            and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
            and "oai_data_id" in request.GET
        ):
            from core_linked_records_app.components.oai_record import (
                api as oai_record_api,
            )

            return JsonResponse(
                {
                    "pid": oai_record_api.get_pid_for_data(
                        request.GET["oai_data_id"], request.user
                    )
                }
            )
        elif (
            "core_federated_search_app" in settings.INSTALLED_APPS
            and "fede_data_id" in request.GET
            and "fede_origin" in request.GET
        ):
            from core_federated_search_app.components.instance import (
                api as instance_api,
            )

            fede_origin_keys = request.GET["fede_origin"].split("&")
            instance_name = fede_origin_keys[1].split("=")[1]
            instance = instance_api.get_by_name(instance_name)

            url_get_data = "%s?data_id=%s" % (
                reverse(
                    "core_linked_record_retrieve_data_pid_url",
                ),
                request.GET["fede_data_id"],
            )
            data_response = oauth2_get_request(
                urljoin(instance.endpoint, url_get_data), instance.access_token
            )
            return JsonResponse(json.loads(data_response.text))
        else:
            return JsonResponse({}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class RetrieveListPID(APIView):
    """Retrieve PIDs for a given list of data IDs."""

    def post(self, request):
        try:
            # FIXME duplicated code with core_explore_common.utils.query.send
            query = query_api.get_by_id(request.POST["query_id"], request.user)
            data_source = query.data_sources[int(request.POST["data_source_index"])]

            # Build serialized query to send to data source
            json_query = {
                "query": query.content,
                "templates": json.dumps(
                    [
                        {"id": str(template.id), "hash": template.hash}
                        for template in query.templates
                    ]
                ),
                "options": json.dumps(data_source.query_options),
                "order_by_field": data_source.order_by_field,
            }

            if (
                getattr(data_source, "capabilities", False)
                and "url_pid" not in data_source.capabilities.keys()
            ):
                return JsonResponse(
                    {"error": "The remote does not have PID capabilities."},
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )

            if data_source.authentication.type == "session":
                response = send_get_request(
                    data_source.capabilities["url_pid"],
                    data=json_query,
                    cookies={"sessionid": request.session.session_key},
                )
            elif data_source.authentication.type == "oauth2":
                response = oauth2_post_request(
                    data_source.capabilities["url_pid"],
                    json_query,
                    data_source.authentication.params["access_token"],
                    session_time_zone=timezone.get_current_timezone(),
                )
            else:
                raise Exception("Unknown authentication type.")

            if response.status_code == 200:
                return JsonResponse(
                    {"pids": [pid for pid in response.json() if pid is not None]},
                    status=response.status_code,
                )
            else:
                return JsonResponse(
                    {
                        "error": "Remote service answered with status code %d."
                        % response.status_code
                    },
                    status=response.status_code,
                )
        except Exception as exception:
            return JsonResponse(
                {"error": str(exception)}, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )



""" Url router for the core linked records app
"""
from django.conf.urls import url

from core_linked_records_app import settings
from core_linked_records_app.rest.providers import views as providers_views
from core_linked_records_app.rest.query import views as query_views
from core_linked_records_app.rest.settings import views as settings_views
from core_linked_records_app.rest.pid import views as pid_views

urlpatterns = [
    url(
        r"^query/local$",
        query_views.ExecuteLocalPIDQueryView.as_view(),
        name="core_linked_records_app_query_local",
    ),
]

if (
    "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
    and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
):
    urlpatterns.append(
        url(
            r"^query/oaipmh$",
            query_views.ExecuteOaiPmhPIDQueryView.as_view(),
            name="core_linked_records_app_query_oaipmh",
        ),
    )

urlpatterns += [
    url(
        r"^settings$",
        settings_views.PidSettings.as_view(),
        name="core_linked_records_app_settings",
    ),
    url(
        r"^retrieve-list-pid",
        pid_views.RetrieveListPID.as_view(),
        name="core_linked_record_retrieve_list_pid_url",
    ),
    url(
        r"^retrieve-data-pid",
        pid_views.RetrieveDataPID.as_view(),
        name="core_linked_record_retrieve_data_pid_url",
    ),
    url(
        r"^(?P<provider>[^/]+)/(?P<record>.*)$",
        providers_views.ProviderRecord.as_view(),
        name="core_linked_records_app_rest_provider_record_view",
    ),
]



"""
"""


def get_dict_value_from_key_list(dictionary, key_list):
    """Browse dictionary content given an ordered list of keys.

    Params:
        dictionary:
        key_list:

    Returns:
    """
    while len(key_list) > 0:
        key = key_list.pop(0)

        if key in dictionary.keys():
            return get_dict_value_from_key_list(dictionary[key], key_list)
        else:
            return None

    return dictionary



""" Apps file for setting linked records when app is ready
"""
import sys

from django.apps import AppConfig

from core_linked_records_app.components.data import watch as data_watch
from core_linked_records_app.settings import AUTO_SET_PID


class LinkedRecordsAppConfig(AppConfig):
    """Core application settings"""

    name = "core_linked_records_app"

    def ready(self):
        """Run when the app is ready

        Returns:

        """
        if "migrate" not in sys.argv:
            if AUTO_SET_PID:
                data_watch.init()



""" Handle system abstract class
"""
from abc import ABC, abstractmethod
from importlib import import_module

from django.urls import reverse

from core_linked_records_app import settings


class AbstractIdProvider(ABC):
    def __init__(self, provider_name, provider_url, username, password):
        self.provider_url = provider_url
        self.local_url = "%s%s" % (
            settings.SERVER_URI,
            reverse(
                "core_linked_records_app_rest_provider_record_view",
                kwargs={"provider": provider_name, "record": ""},
            ),
        )
        self.local_url = self.local_url[:-1]  # Removing the final /
        self.auth_token = self.encode_token(username, password)

    @abstractmethod
    def encode_token(self, username, password):
        raise NotImplementedError()

    @abstractmethod
    def get(self, record):
        raise NotImplementedError()

    @abstractmethod
    def create(self, prefix, record=None):
        raise NotImplementedError()

    @abstractmethod
    def update(self, record):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, record):
        raise NotImplementedError()


class ProviderManager(object):
    """Manage provider instances from a given provider name"""

    def __init__(self):
        self._provider_instances = {
            provider_name: None for provider_name in settings.ID_PROVIDER_SYSTEMS.keys()
        }

    def get(self, provider_name):
        if self._provider_instances[provider_name] is None:
            # Retrieve class name and module path
            id_provider_classpath = settings.ID_PROVIDER_SYSTEMS[provider_name][
                "class"
            ].split(".")
            id_provider_classname = id_provider_classpath[-1]
            id_provider_modpath = ".".join(id_provider_classpath[:-1])

            # Import module and class
            id_provider_module = import_module(id_provider_modpath)
            id_provider_class = getattr(id_provider_module, id_provider_classname)

            # Initialize handel system instance
            self._provider_instances[provider_name] = id_provider_class(
                provider_name, *settings.ID_PROVIDER_SYSTEMS[provider_name]["args"]
            )

        return self._provider_instances[provider_name]

    def find_provider_from_pid(self, pid):
        for provider_name in self._provider_instances.keys():
            provider = self.get(provider_name)

            if pid.startswith(provider.provider_url):
                return provider_name

        return None



""" Linked records REST views
"""
import json
import logging
import re

from rest_framework import status
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.views import APIView

from core_linked_records_app import settings
from core_linked_records_app.components.data.api import get_data_by_pid
from core_linked_records_app.rest.data.renderers.data_html_user_renderer import (
    DataHtmlUserRenderer,
)
from core_linked_records_app.rest.data.renderers.data_xml_renderer import (
    DataXmlRenderer,
)
from core_linked_records_app.utils.providers import ProviderManager
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.data.serializers import DataSerializer

LOGGER = logging.getLogger(__name__)


class ProviderRecord(APIView):
    parser_classes = (JSONParser,)
    renderer_classes = (DataHtmlUserRenderer, JSONRenderer, DataXmlRenderer)

    def __init__(self):
        self.provider_manager = ProviderManager()
        super().__init__()

    def post(self, request, provider, record):
        """Create a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        # Parse record entry to split between prefix and record ID.
        prefix_record_list = record.split("/")

        if len(prefix_record_list) == 1:  # No record provided
            prefix = prefix_record_list[0]
            record = None
        elif prefix_record_list[-1] == "":  # Only prefix provided
            prefix = "/".join(prefix_record_list[:-1])
            record = None
        else:  # Prefix and record provided
            prefix = "/".join(prefix_record_list[:-1])
            record = prefix_record_list[-1]

        # Assign default prefix if the prefix is undefined or not in the list of
        # authorized ones.
        if prefix == "" or prefix not in settings.ID_PROVIDER_PREFIXES:
            return Response(
                {
                    "record": "/".join(prefix_record_list),
                    "url": request.build_absolute_uri("?"),
                    "message": "Prefix not existing",
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        if (
            record is not None
            and record != ""
            and re.match(r"^(%s|)$" % settings.PID_FORMAT, record) is None
        ):
            return Response(
                {
                    "record": "/".join(prefix_record_list),
                    "url": request.build_absolute_uri("?"),
                    "message": "Invalid record format",
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        id_provider = self.provider_manager.get(provider)
        provider_response = id_provider.create(prefix, record)

        provider_content = json.loads(provider_response.content)
        return Response(provider_content, status=provider_response.status_code)

    def put(self, request, provider, record):
        """Update the value of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self.provider_manager.get(provider)
        provider_response = id_provider.update(record)

        provider_content = json.loads(provider_response.content)

        return Response(provider_content, status=provider_response.status_code)

    def get(self, request, provider, record):
        """Retrieve the local data of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self.provider_manager.get(provider)
        provider_response = id_provider.get(record)

        try:
            query_result = get_data_by_pid(
                json.loads(provider_response.content)["url"], request.user
            )
            return Response(
                DataSerializer(query_result).data, status=status.HTTP_200_OK
            )
        except DoesNotExist:
            content = {
                "status": "error",
                "code": status.HTTP_404_NOT_FOUND,
                "message": "No data with specified handle found",
            }
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as ex:
            content = {
                "status": "error",
                "code": status.HTTP_500_INTERNAL_SERVER_ERROR,
                "message": str(ex),
            }
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def delete(self, request, provider, record):
        """Delete a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self.provider_manager.get(provider)
        provider_response = id_provider.delete(record)

        provider_content = json.loads(provider_response.content)

        return Response(provider_content, status=provider_response.status_code)



""" Core linked records app settings

Settings with the following syntax can be overwritten at the project level:
SETTING_NAME = getattr(settings, "SETTING_NAME", "Default Value")
"""
from django.conf import settings

if not settings.configured:
    settings.configure()

INSTALLED_APPS = getattr(
    settings,
    "INSTALLED_APPS",
    [],
)
SERVER_URI = getattr(settings, "SERVER_URI", "http://localhost")

PID_XPATH = getattr(settings, "PID_XPATH", "Resource.@localid")

PID_FORMAT = getattr(settings, "PID_FORMAT", r"[a-zA-Z0-9_\-]+")

ID_PROVIDER_SYSTEMS = getattr(
    settings,
    "ID_PROVIDER_SYSTEMS",
    {
        "local": {
            "class": "core_linked_records_app.utils.providers.local.LocalIdProvider",
            "args": [SERVER_URI],
        }
    },
)

ID_PROVIDER_PREFIXES = getattr(settings, "ID_PROVIDER_PREFIXES", ["cdcs"])

ID_PROVIDER_PREFIX_DEFAULT = getattr(
    settings, "ID_PROVIDER_PREFIX_DEFAULT", ID_PROVIDER_PREFIXES[0]
)

AUTO_SET_PID = getattr(settings, "AUTO_SET_PID", False)



""" Signals to trigger before Data save
"""
import logging
from os.path import join

from rest_framework import status

from core_linked_records_app import settings
from core_linked_records_app.settings import (
    PID_XPATH,
)
from core_linked_records_app.system import api as system_api
from core_linked_records_app.utils.providers import ProviderManager
from core_linked_records_app.utils.xml import (
    get_xpath_from_dot_notation,
    get_xpath_with_target_namespace,
    get_value_at_xpath,
    set_value_at_xpath,
)
from core_main_app.commons import exceptions
from core_main_app.components.data.models import Data
from core_main_app.utils.requests_utils.requests_utils import send_post_request
from signals_utils.signals.mongo import signals, connector
from xml_utils.xpath import create_tree_from_xpath
from xml_utils.xsd_tree.xsd_tree import XSDTree

LOGGER = logging.getLogger(__name__)


def init():
    """Connect to Data object events."""
    connector.connect(set_data_pid, signals.pre_save, sender=Data)


def set_data_pid(sender, document, **kwargs):
    """Set the PID in the XML field specified in the settings. If the PID
    already exists and is valid, it is not reset.

    Params:
        sender:
        document:
        kwargs:

    Returns:
    """
    pid_xpath = get_xpath_from_dot_notation(PID_XPATH)

    # Retrieve namespaces
    xml_tree = XSDTree.build_tree(document.xml_content)
    pid_xpath, namespaces = get_xpath_with_target_namespace(
        pid_xpath, document.template.content
    )

    try:  # Get the PID from the `pid_xpath` value
        document_pid = get_value_at_xpath(xml_tree, pid_xpath, namespaces)
        if type(document_pid) == str and document_pid.endswith(
            "/"
        ):  # Cleanup PID if it ends with a '/'
            document_pid = document_pid[:-1]
    except AssertionError:  # PID XPath not found in document
        try:  # Try to create the element at the given PID
            # Import libs that need to wait for apps to be ready
            from core_main_app.components.data import api as data_api

            modified_xml_tree = create_tree_from_xpath(pid_xpath, xml_tree, namespaces)
            set_value_at_xpath(
                modified_xml_tree, pid_xpath, "http://sample_pid.org", namespaces
            )
            document.xml_content = XSDTree.tostring(modified_xml_tree)
            data_api.check_xml_file_is_valid(document)

            xml_tree = modified_xml_tree
            document_pid = None
        except Exception as exc:  # Cannot create PID at given XPath
            LOGGER.warning("Cannot create PID at %s: %s" % (pid_xpath, str(exc)))
            return

    # Identify provider name for registration and ensure the PID has not been
    # already defined in another document.
    provider_manager = ProviderManager()

    if document_pid is None or document_pid == "":  # PID field left blank
        # Select the default provider if no PID has been chosen.
        provider_name = list(settings.ID_PROVIDER_SYSTEMS.keys())[0]
        document_pid = join(
            provider_manager.get(provider_name).provider_url,
            settings.ID_PROVIDER_PREFIX_DEFAULT,
        )
    else:  # PID specified in document.
        # Check that the PID is not defined for a document other than the current
        # document.
        if system_api.is_pid_defined(document_pid) and (
            document.pk is None
            or not system_api.is_pid_defined_for_document(document_pid, document.pk)
        ):
            raise exceptions.ModelError("PID already defined for another document")

        provider_name = provider_manager.find_provider_from_pid(document_pid)

    # PID specified but not matching any possible provider URLs for the
    # generation.
    if provider_name is None:
        raise exceptions.ModelError("Invalid PID provided")

    provider = provider_manager.get(provider_name)
    registration_url = document_pid.replace(provider.provider_url, provider.local_url)

    document_pid_response = send_post_request("%s?format=json" % registration_url)

    if document_pid_response.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR:
        default_error_message = "An error occured while creating the PID"
        try:
            raise exceptions.ModelError(
                document_pid_response.json().get("message", default_error_message)
            )
        except ValueError:  # If the response is not JSON parsable
            raise exceptions.ModelError(default_error_message)

    if (
        document_pid_response.status_code != status.HTTP_201_CREATED
        and document_pid_response.status_code != status.HTTP_200_OK
        and system_api.get_data_by_pid(document_pid).pk != document.pk
    ):
        raise exceptions.ModelError("Invalid PID provided")

    # FIXME assert the value is not changed when saving
    document_pid = document_pid_response.json()["url"]

    # Set the document PID into XML data and update `xml_content`
    set_value_at_xpath(xml_tree, pid_xpath, document_pid, namespaces)
    document.xml_content = XSDTree.tostring(xml_tree)

    # Update the whole document with the updated XML content
    document.convert_to_file()
    document.convert_to_dict()



""" Local resolver API
"""
from core_linked_records_app.settings import PID_XPATH
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_oaipmh_harvester_app.components.oai_record import api as oai_record_data


def get_pid_for_data(oai_record_id, user):
    """Retrieve PID matching the document ID provided.

    Args:
        oai_record_id:

    Returns:
    """
    # Retrieve the document passed as input and extra the PID field.
    data = oai_record_data.get_by_id(oai_record_id, user)

    # Return PID value from the document and the PID_XPATH
    return get_dict_value_from_key_list(data["dict_content"], PID_XPATH.split("."))



""" Rest API views to retrieve PID settings
"""
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from core_linked_records_app import settings


class PidSettings(APIView):
    """Retrieve PID settings"""

    def get(self, request):
        return Response(
            {
                "xpath": settings.PID_XPATH,
                "format": settings.PID_FORMAT,
                "systems": list(settings.ID_PROVIDER_SYSTEMS.keys()),
                "prefixes": settings.ID_PROVIDER_PREFIXES,
                "auto_set": settings.AUTO_SET_PID,
            },
            status=status.HTTP_200_OK,
        )



""" Local record API
"""
from mongoengine import NotUniqueError as MongoNotUniqueError

from core_linked_records_app.components.local_id.models import LocalId
from core_main_app.commons.exceptions import NotUniqueError


def get_by_name(record_name):
    """Retrieve the record by name.

    Args:
        record_name:

    Returns:
    """
    record_name = record_name.lower()
    return LocalId.get_by_name(record_name)


def insert(record):
    """Insert the record in the collection.

    Args:
        record:

    Returns:
    """
    try:
        record.record_name = record.record_name.lower()
        return record.save()
    except MongoNotUniqueError as e:
        raise NotUniqueError(e)


def delete(record):
    """Delete the record.

    Args:
        record:

    Returns:

    """
    return record.delete()
